import math
from typing import Union
from .parameter_range import ParameterRange


class Parameter:

    def __init__(
            self,
            parameter_name: str,
            minimum: Union[int, float]=None,
            maximum: Union[int, float]=None,
            feed_forward: bool=False
        ) -> None:
        self.parameter_name = parameter_name
        self.feed_forward = feed_forward
        self.minimum = minimum
        self.maximum = maximum

        if minimum is None or minimum <= 0 and self.feed_forward:
            minimum = 0.5

        if maximum is None or maximum <= 0 and self.feed_forward:
            maximum = math.ceil(minimum) * 2

        self.range = ParameterRange(
            minimum_range=minimum,
            maximum_range=maximum
        )

        assert self.maximum > self.minimum, f"Err. - maximum parameter range value for optimization parameter {self.parameter_name} must be greater than minimum parameter range value."

        if self.feed_forward:
            assert self.minimum >= 0.1, f'Err. - Range multiplier for feed-forward optimization parameter {self.parameter_name} cannot exceed minimum of 0.1.'
            assert self.maximum <= 10, f'Err. - Range multiplier for feed-forward optimization parameter {self.parameter_name} cannot exceed maximum of 10.'
            


from typing import (
    List, 
    Union, 
    Dict, 
    Any, 
    Type, 
    Callable, 
    Awaitable, 
    Tuple
)
from hedra.core.engines.types.common.base_action import BaseAction
from hedra.core.engines.types.common.base_engine import BaseEngine
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.hook_metadata import HookMetadata


class TaskHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        *names: Tuple[str, ...],
        weight: int=1, 
        order: int=1, 
        skip: bool=False,
        metadata: Dict[str, Union[str, int]]={}
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            order=order,
            skip=skip,
            hook_type=HookType.TASK
        )

        self.names = list(set(names))
        self.call: Type[self._call] = self._call
        self.session: BaseEngine = None
        self.action: BaseAction = None
        self.order = order
        self.before: List[Any] = []
        self.after: List[Any] = []
        self.is_notifier = False
        self.is_listener = False
        self.checks: List[Any] = []
        self.channels: List[Any] = []
        self.notifiers: List[Any] = []
        self.listeners: List[Any] = []
        self.metadata = HookMetadata(
            weight=weight,
            order=order,
            **metadata
        )

    def copy(self):
        task_hook = TaskHook(
            self.name,
            self.shortname,
            self._call,
            weight=self.metadata.weight,
            order=self.order,
            skip=self.skip,
            metadata={
                **self.metadata.copy()
            }
        )

        task_hook.checks = list(self.checks)
        task_hook.stage = self.stage

        return task_hook
    
    async def call(self, *args, **kwargs):
        return await self._call(*args, **kwargs)




from typing import List
from .connection import WebsocketConnection


class Pool:

    __slots__ = (
        'size',
        'connections',
        'reset_connections'
    )

    def __init__(self, size: int, reset_connections: bool = False) -> None:
        self.size = size
        self.connections: List[WebsocketConnection] = []
        self.reset_connections = reset_connections

    def create_pool(self) -> None:
        for _ in range(self.size):
            self.connections.append(
                WebsocketConnection(self.reset_connections)
            )

    async def close(self):
        for connection in self.connections:
            await connection.close()



import uuid
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.grpc import (
    GRPCAction,
    MercuryGRPCClient
)
from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.grpc.protobuf_registry import protobuf_registry
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.hooks.types.base.simple_context import SimpleContext
from hedra.data.parsers.parser_types.common.base_parser import BaseParser
from hedra.data.parsers.parser_types.common.parsing import (
    normalize_headers,
    parse_data,
    parse_tags
)
from typing import Any, Coroutine, Dict
from .grpc_action_validator import GRPCActionValidator
from .grpc_options_validator import GRPCOptionsValidator


class GRPCActionParser(BaseParser):

    def __init__(
        self,
        config: Config,
        options: Dict[str, Any]={}
    ) -> None:
        super().__init__(
            GRPCActionParser.__name__,
            config,
            RequestTypes.GRPC,
            options
        )

        self.grpc_options = GRPCOptionsValidator(
            protobuf_map=protobuf_registry
        )

        self.protobuf_map = self.grpc_options.protobuf_map

    async def parse(
        self, 
        action_data: Dict[str, Any],
        stage: str
    ) -> Coroutine[Any, Any, Coroutine[Any, Any, ActionHook]]:
        
        action_name = action_data.get('name')
        normalized_headers = normalize_headers(action_data)
        tags_data = parse_tags(action_data)

        protobuf = self.grpc_options.protobuf_map[action_name].ParseFromString(
            action_data.get('data')
        )

        generator_action = GRPCActionValidator(**{
            **action_data,
            'headers': normalized_headers,
            'data': protobuf,
            'tags': tags_data
        })

        action = GRPCAction(
            generator_action.name,
            generator_action.url,
            method=generator_action.method,
            headers=generator_action.headers,
            data=generator_action.data,
            user=generator_action.user,
            tags=[
                tag.dict() for tag in generator_action.tags
            ]
        )

        session = MercuryGRPCClient(
            concurrency=self.config.batch_size,
            timeouts=self.timeouts,
            reset_connections=self.config.reset_connections,
            tracing_session=self.config.tracing
        )

        await session.prepare(action)
        
        hook = ActionHook(
            f'{stage}.{generator_action.name}',
            generator_action.name,
            None,
            order=generator_action.order,
            weight=generator_action.weight,
            metadata={
                'user': generator_action.user,
                'tags': generator_action.tags   
            }
        )

        hook.session = session
        hook.action = action
        hook.stage = stage
        hook.context = SimpleContext()
        hook.hook_id = uuid.uuid4()

        hook.metadata.order = generator_action.order
        hook.metadata.weight = generator_action.weight
        hook.metadata.tags = generator_action.tags
        hook.metadata.user = generator_action.user


        return hook






from .types import (
    AWSLambdaConfig,
    AWSTimestreamConfig,
    BigQueryConfig,
    BigTableConfig,
    CassandraConfig,
    CloudwatchConfig,
    CosmosDBConfig,
    CSVConfig,
    DatadogConfig,
    DogStatsDConfig,
    GoogleCloudStorageConfig,
    GraphiteConfig,
    HoneycombConfig,
    InfluxDBConfig,
    JSONConfig,
    KafkaConfig,
    MongoDBConfig,
    MySQLConfig,
    NetdataConfig,
    NewRelicConfig,
    PostgresConfig,
    PrometheusConfig,
    RedisConfig,
    S3Config,
    SnowflakeConfig,
    SQLiteConfig,
    StatsDConfig,
    TelegrafConfig,
    TelegrafStatsDConfig,
    TimescaleDBConfig,
    XML,
    XMLConfig
    
)

from .reporter import (
    Reporter,
    ReporterType
)


from hedra.reporting.metric.metric_types import MetricType
from hedra.reporting.tags import Tag
from pydantic import (
    BaseModel, 
    StrictStr, 
    StrictBool, 
    StrictInt, 
    StrictFloat
)
from typing import (
    Dict, 
    Union,
    List
)


class MutationSummary(BaseModel):
    mutation_name: StrictStr
    mutation_experiment_name: StrictStr
    mutation_variant_name: StrictStr
    mutation_chance: StrictFloat
    mutation_targets: StrictStr
    mutation_type: StrictStr

    @property
    def stats(self) -> Dict[str, Union[int, float]]:
        return {
            'mutation_chance': self.mutation_chance
        }
    
    @property
    def record(self) -> Dict[str, Union[str, bool, int, float]]:
        return {
            'mutation_name': self.mutation_name,
            'mutation_experiment_name': self.mutation_experiment_name,
            'mutation_variant_name': self.mutation_variant_name,
            'mutation_targets': self.mutation_targets,
            'mutation_type': self.mutation_type,
            **self.stats
        }
    
    @property
    def types_map(self):
        return {
            'mutation_chance': MetricType.SAMPLE
        }
    
    @property
    def tags(self):
        tag_fields = {
            'mutation_name': self.mutation_name,
            'mutation_experiment_name': self.mutation_experiment_name,
            'mutation_variant_name': self.mutation_variant_name,
            'mutation_targets': self.mutation_targets,
            'mutation_type': self.mutation_type,
        }

        return [
            Tag(
                tag_field_name,
                tag_field_value
            ) for tag_field_name, tag_field_value in tag_fields.items()
        ]


class VariantSummary(BaseModel):
    variant_name: StrictStr
    variant_experiment: StrictStr
    variant_weight: StrictFloat
    variant_distribution: Union[StrictStr, None]
    variant_distribution_interval: Union[StrictInt, StrictFloat]
    variant_completed: StrictInt
    variant_succeeded: StrictInt
    variant_failed: StrictInt
    variant_actions_per_second: Union[List[StrictFloat], StrictFloat]
    variant_mutation_summaries: Dict[str, MutationSummary]
    variant_ratio_completed: Union[StrictFloat, None]
    variant_ratio_succeeded: Union[StrictFloat, None]
    variant_ratio_failed: Union[StrictFloat, None]
    variant_ratio_aps: Union[StrictFloat, None]

    @property
    def stats(self) -> Dict[str, Union[int, float]]:
        return {
            'variant_weight': self.variant_weight,
            'variant_distribution_interval': self.variant_distribution_interval,
            'variant_completed': self.variant_completed,
            'variant_succeeded': self.variant_succeeded,
            'variant_failed': self.variant_failed,
            'variant_actions_per_second': self.variant_actions_per_second,
            'variant_ratio_completed': self.variant_ratio_completed,
            'variant_ratio_succeeded': self.variant_ratio_succeeded,
            'variant_ratio_failed': self.variant_ratio_failed,
            'variant_ratio_aps': self.variant_ratio_aps
        }
    
    @property
    def record(self) -> Dict[str, Union[str, bool, int, float]]:
        return {
            'variant_name': self.variant_name,
            'variant_experiment': self.variant_experiment,
            'variant_distribution': self.variant_distribution,
            **self.stats
        }
    
    @property
    def types_map(self):
        return {
            'variant_weight': MetricType.SAMPLE,
            'variant_distribution_interval': MetricType.SAMPLE,
            'variant_completed': MetricType.COUNT,
            'variant_succeeded': MetricType.COUNT,
            'variant_failed': MetricType.COUNT,
            'variant_actions_per_second': MetricType.RATE,
            'variant_ratio_completed': MetricType.SAMPLE,
            'variant_ratio_succeeded': MetricType.SAMPLE,
            'variant_ratio_failed': MetricType.SAMPLE,
            'variant_ratio_aps': MetricType.SAMPLE
        }

    @property
    def tags(self):
        tag_fields = {
            'variant_name': self.variant_name,
            'variant_experiment': self.variant_experiment,
            'variant_distribution': self.variant_distribution,
        }

        return [
            Tag(
                tag_field_name,
                tag_field_value
            ) for tag_field_name, tag_field_value in tag_fields.items()
        ]


class ExperimentSummary(BaseModel):
    experiment_name: StrictStr
    experiment_randomized: StrictBool
    experiment_completed: StrictInt
    experiment_succeeded: StrictInt
    experiment_failed: StrictInt
    experiment_median_aps: StrictFloat
    experiment_variant_summaries: Dict[str, VariantSummary]

    @property
    def stats(self) -> Dict[str, Union[int, float]]:
        return {
            'experiment_completed': self.experiment_completed,
            'experiment_succeeded': self.experiment_succeeded,
            'experiment_failed': self.experiment_failed,
            'experiment_median_aps': self.experiment_median_aps
        }
    
    @property
    def record(self) -> Dict[str, Union[str, bool, int, float]]:
        return {
            'experiment_name': self.experiment_name,
            'experiment_randomized': self.experiment_randomized,
            **self.stats
        }
    
    @property
    def types_map(self):
        return {
            'experiment_completed': MetricType.COUNT,
            'experiment_succeeded': MetricType.COUNT,
            'experiment_failed': MetricType.COUNT,
            'experiment_median_aps': MetricType.SAMPLE
        }
    
    @property
    def tags(self):
        tag_fields = {
            'experiment_name': self.experiment_name,
            'experiment_randomized': self.experiment_randomized,
        }

        return [
            Tag(
                tag_field_name,
                tag_field_value
            ) for tag_field_name, tag_field_value in tag_fields.items()
        ]


import asyncio
import uuid
from typing import (
    List,
    Dict,
    Any,
    Union,
    Coroutine
)
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .cosmos_connector_config import CosmosDBConnectorConfig

try:
    from azure.cosmos.aio import CosmosClient
    from azure.cosmos import DatabaseProxy
    has_connector = True
except Exception:
    CosmosClient = None
    DatabaseProxy = None
    has_connector = False


class CosmosDBConnector:
    connector_type=ConnectorType.CosmosDB

    def __init__(
        self, config: CosmosDBConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.account_uri = config.account_uri
        self.account_key = config.account_key

        self.database_name = config.database
        self.container_name = config.container_name
        self.partition_key = config.partition_key

        self.analytics_ttl = config.analytics_ttl

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.stage = stage
        self.parser_config = parser_config

        self.logger = HedraLogger()
        self.logger.initialize()

        self.container = None
        
        self.client = None
        self.database: Union[DatabaseProxy, None] = None
        self.parser = Parser()

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to CosmosDB')

        self.client = CosmosClient(
            self.account_uri,
            credential=self.account_key
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to CosmosDB')
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Database - {self.database_name} - if not exists')
        self.database = self.client.get_database_client(self.database_name)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Database - {self.database_name}')

        self.container = self.database.get_container_client(self.container_name)
    
    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)
    
    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        actions = await self.load_data()

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data()

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        return [
            record async for record in self.container.read_all_items(
                max_item_count=options.get('max_item_count'),
            )
        ]
            
    
    async def close(self):
        await self.client.close()



from pydantic import BaseModel
from hedra.data.connectors.common.connector_type import ConnectorType


class CosmosDBConnectorConfig(BaseModel):
    account_uri: str
    account_key: str
    database: str
    container_name: str
    partition_key: str
    analytics_ttl: int=0
    connector_type: ConnectorType=ConnectorType.CosmosDB


from hedra.core.hooks.types.base.hook import Hook
from hedra.core.hooks.types.base.hook_type import HookType


class HookSetupError(Exception):

    def __init__(self, hook: Hook, hook_type: HookType, message: str) -> None:

        hook_type = hook_type.name.lower()
        super().__init__(
            f'Hook Error - @{hook_type} hook {hook.shortname} from stage {hook.stage}\n{message}'
        )




from pydantic import (
    BaseModel, 
    StrictBool, 
    StrictFloat, 
    StrictInt, 
    StrictStr
)
from typing import List, Dict, Union
from .experiment_metrics_set_types import (
    ExperimentSummary,
    VariantSummary,
    MutationSummary
)


class ExperimentMetricsCollection(BaseModel):
    experiment: Dict[str, Union[StrictBool, StrictFloat, StrictInt, StrictStr]]
    variants: List[Dict[str, Union[StrictBool, StrictFloat, StrictInt, StrictStr]]]
    mutations: List[Dict[str, Union[StrictBool, StrictFloat, StrictInt, StrictStr]]]

    experiment_summary: ExperimentSummary
    variant_summaries: List[VariantSummary]
    mutation_summaries: List[MutationSummary]


class ExperimentMetricsCollectionSet(BaseModel):
    experiments_metrics_fields: List[str]
    variants_metrics_fields: List[str]
    mutations_metrics_fields: List[str]
    experiments: List[Dict[str, Union[StrictBool, StrictFloat, StrictInt, StrictStr]]]
    variants: List[Dict[str, Union[StrictBool, StrictFloat, StrictInt, StrictStr]]]
    mutations: List[Dict[str, Union[StrictBool, StrictFloat, StrictInt, StrictStr]]]

    experiment_summaries: List[ExperimentSummary]
    variant_summaries: List[VariantSummary]
    mutation_summaries: List[MutationSummary]
    


import asyncio
import uuid
import functools
import json
import signal
import psutil
from datetime import datetime
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import MetricsSet
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from concurrent.futures import ThreadPoolExecutor
from .s3_config import S3Config


try:
    import boto3
    has_connector = True

except Exception:
    boto3 = None
    has_connector = False


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


class S3:

    def __init__(self, config: S3Config) -> None:
        self.aws_access_key_id = config.aws_access_key_id
        self.aws_secret_access_key = config.aws_secret_access_key
        self.region_name = config.region_name
        self.buckets_namespace = config.buckets_namespace

        self.events_bucket_name = config.events_bucket
        self.metrics_bucket_name = config.metrics_bucket
        self.streams_bucket_name = config.streams_bucket

        self.experiments_bucket_name = config.experiments_bucket
        self.variants_bucket_name = f'{config.experiments_bucket}_variants'
        self.mutations_bucket_name = f'{config.experiments_bucket}_mutations'

        self.shared_metrics_bucket_name = f'{config.metrics_bucket}_shared'
        self.errors_bucket_name = f'{config.metrics_bucket}_errors'
        self.custom_metrics_bucket_name = f'{config.metrics_bucket}_custom'

        self.session_system_metrics_bucket_name = f'{config.system_metrics_bucket}_session'
        self.stage_system_metrics_bucket_name = f'{config.system_metrics_bucket}_stage'

        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self.client = None
        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to AWS S3 - Region: {self.region_name}')
        
        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop
                )
            )

        self.client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                's3',
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                region_name=self.region_name
            )
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to AWS S3 - Region: {self.region_name}')
    
    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):
        
        session_system_metrics_bucket_name = f'{self.buckets_namespace}-{self.session_system_metrics_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Bucket - {session_system_metrics_bucket_name}')
        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Session System Metrics Bucket - {session_system_metrics_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=session_system_metrics_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Session System Metrics Bucket - {session_system_metrics_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Session System Metrics Bucket - {session_system_metrics_bucket_name}')

        metrics_sets: List[SessionMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.system_metrics_set_id}')
            for monitor_metrics in metrics_set.session_cpu_metrics.values():
                metrics_sets.append(monitor_metrics)
                
            for  monitor_metrics in metrics_set.session_memory_metrics.values():
                metrics_sets.append(monitor_metrics)

        for metrics_set in metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics Set - {metrics_set.name}:{metrics_set.group}')

            session_system_metrics_set_id = uuid.uuid4()

            metrics_set_set_key = f'{metrics_set.name}_{metrics_set.group}_{session_system_metrics_set_id}'
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=session_system_metrics_bucket_name,
                    Key=metrics_set_set_key,
                    Body=json.dumps(metrics_set.record)
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to Bucket - {session_system_metrics_bucket_name}')
    
    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):
        
        stage_system_metrics_bucket_name = f'{self.buckets_namespace}-{self.stage_system_metrics_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Bucket - {stage_system_metrics_bucket_name}')
        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Stage System Metrics Bucket - {stage_system_metrics_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=stage_system_metrics_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Stage System Metrics Bucket - {stage_system_metrics_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Stage System Metrics Bucket - {stage_system_metrics_bucket_name}')

        metrics_sets: List[SystemMetricsCollection] = []
        
        for metrics_set in system_metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.system_metrics_set_id}')
            
            cpu_metrics = metrics_set.cpu
            memory_metrics = metrics_set.memory

            for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                for monitor_metrics in stage_cpu_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                for monitor_metrics in stage_memory_metrics.values():
                    metrics_sets.append(monitor_metrics)

                stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                
                if stage_mb_per_vu_metrics:
                    metrics_sets.append(stage_mb_per_vu_metrics)

        for metrics_set in metrics_sets:

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics Set - {metrics_set.name}:{metrics_set.group}')

            stage_system_metrics_set_id = uuid.uuid4()

            metrics_set_set_key = f'{metrics_set.name}_{metrics_set.group}_{stage_system_metrics_set_id}'
            
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=stage_system_metrics_bucket_name,
                    Key=metrics_set_set_key,
                    Body=json.dumps(metrics_set.record)
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to Bucket - {stage_system_metrics_bucket_name}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        streams_bucket_name = f'{self.buckets_namespace}-{self.streams_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Bucket - {streams_bucket_name}')
        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Streams Bucket - {streams_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=streams_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Streams Bucket - {streams_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Streams Bucket - {streams_bucket_name}')

        for stage_name, stream in stream_metrics.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams Set - {stage_name}:{stream.stream_set_id}')

            for group_name, group in stream.grouped.items():
                streams_set_key = f'{stage_name}_{group_name}_{stream.stream_set_id}'
                
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.put_object,
                        Bucket=streams_bucket_name,
                        Key=streams_set_key,
                        Body=json.dumps({
                            **group.record,
                            'group': group_name,
                            'stage': stage_name,
                            'name': f'{stage_name}_streams'
                        })
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to Bucket - {streams_bucket_name}')

    async def submit_experiments(self, experiment_metrics: ExperimentMetricsCollectionSet):

        experiments_bucket_name = f'{self.buckets_namespace}-{self.experiments_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Bucket - {experiments_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Experiments Bucket - {experiments_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=experiments_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Experiments Bucket - {experiments_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Experiments Bucket - {experiments_bucket_name}')
        
        for experiment in experiment_metrics.experiment_summaries:
            
            experiment_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiment - {experiment.experiment_name}:{experiment_id}')
      
            experiment_key = f'{experiment.experiment_name}_{experiment_id}'

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=experiments_bucket_name,
                    Key=experiment_key,
                    Body=json.dumps(experiment.record)
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to Bucket - {experiments_bucket_name}')

    async def submit_variants(self, experiment_metrics: ExperimentMetricsCollectionSet):

        variants_bucket_name = f'{self.buckets_namespace}-{self.variants_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Bucket - {variants_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Variants Bucket - {variants_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=variants_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Variants Bucket - {variants_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Variants Bucket - {variants_bucket_name}')
        
        for variant in experiment_metrics.variant_summaries:
            
            variant_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variant - {variant.variant_name}:{variant_id}')
      
            variant_key = f'{variant.variant_name}_{variant_id}'

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=variants_bucket_name,
                    Key=variant_key,
                    Body=json.dumps(variant.record)
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to Bucket - {variants_bucket_name}')

    async def submit_mutations(self, experiment_metrics: ExperimentMetricsCollectionSet):

        mutations_bucket_name = f'{self.buckets_namespace}-{self.mutations_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Bucket - {mutations_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Mutations Bucket - {mutations_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=mutations_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Mutations Bucket - {mutations_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Mutations Bucket - {mutations_bucket_name}')
        
        for mutation in experiment_metrics.mutation_summaries:
            
            mutation_id = uuid.uuid4()

            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutation - {mutation.mutation_name}:{mutation_id}')
      
            mutation_key = f'{mutation.mutation_name}_{mutation_id}'

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=mutations_bucket_name,
                    Key=mutation_key,
                    Body=json.dumps(mutation.record)
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to Bucket - {mutations_bucket_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):
        
        events_bucket_name = f'{self.buckets_namespace}-{self.events_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Bucket - {events_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Events Bucket - {events_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=events_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Events Bucket - {events_bucket_name} - if not exists')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Events Bucket - {events_bucket_name}')


        for event in events:
            event_key = f'{event.name}_{event.event_id}'

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=events_bucket_name,
                    Key=event_key,
                    Body=json.dumps(event.record)
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Bucket - {events_bucket_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        shared_metrics_bucket_name = f'{self.buckets_namespace}-{self.shared_metrics_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Bucket - {shared_metrics_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Shared Metrics Bucket - {shared_metrics_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=shared_metrics_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Shared Metrics Bucket - {shared_metrics_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Shared Metrics Bucket - {shared_metrics_bucket_name}')
        
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
      
            metrics_set_key = f'{metrics_set.name}_{metrics_set.metrics_set_id}'

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=shared_metrics_bucket_name,
                    Key=metrics_set_key,
                    Body=json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        **metrics_set.common_stats
                    })
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Bucket - {shared_metrics_bucket_name}')
    
    async def submit_metrics(self, metrics: List[MetricsSet]):

        metrics_bucket_name = f'{self.buckets_namespace}-{self.metrics_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Bucket - {metrics_bucket_name}')
        
        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Metrics Bucket - {metrics_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=metrics_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Metrics Bucket - {metrics_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Metrics Bucket - {metrics_bucket_name}')

        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                metrics_set_key = f'{metrics_set.name}_{group_name}_{group.metrics_group_id}'
                
                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.put_object,
                        Bucket=metrics_bucket_name,
                        Key=metrics_set_key,
                        Body=json.dumps({
                            **group.record,
                            'group': group_name
                        })
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Bucket - {metrics_bucket_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Custom Metrics Bucket - {self.custom_metrics_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=self.custom_metrics_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Custom Metrics Bucket - {self.custom_metrics_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Custom Metrics Bucket - {self.custom_metrics_bucket_name}')


        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
               
            custom_metric_key = f'{metrics_set.name}_custom_{metrics_set.metrics_set_id}'

            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.put_object,
                    Bucket=self.custom_metrics_bucket_name,
                    Key=custom_metric_key,
                    Body=json.dumps({
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': 'custom',
                        **{
                            custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                        }
                    })
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to Bucket - {self.custom_metrics_bucket_name}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        errors_bucket_name = f'{self.buckets_namespace}-{self.errors_bucket_name}'
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Bucket - {errors_bucket_name}')

        try:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Error Metrics Bucket - {errors_bucket_name} - if not exists')
            await self._loop.run_in_executor(
                self._executor,
                functools.partial(
                    self.client.create_bucket,
                    Bucket=errors_bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': self.region_name
                    }
                )
            )

            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created Error Metrics Bucket - {errors_bucket_name}')

        except Exception:
            await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Skipping creation of Error Metrics Bucket - {errors_bucket_name}')

        for metrics_set in metrics_sets:
            for error in metrics_set.errors:
                timestamp = int(datetime.now().timestamp())
                metric_key = f'{metrics_set.name}-{timestamp}'

                await self._loop.run_in_executor(
                    self._executor,
                    functools.partial(
                        self.client.put_object,
                        Bucket=errors_bucket_name,
                        Key=metric_key,
                        Body=json.dumps({
                            'metrics_name': metrics_set.name,
                            'metrics_stage': metrics_set.stage,
                            'error_message': error.get('message'),
                            'error_count': error.get('count')
                        })
                    )
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Bucket - {errors_bucket_name}')

    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        self._executor.shutdown()


from .differential_evolution_optimizer import DifferentialEvolutionOptimizer
from .dual_annealing_optimizer import DualAnnealingOptimizer
from .point_optimizer import PointOptimizer
from .shg_optimizer import SHGOptimizer


import asyncio
import functools
import psutil
import signal
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from .exceptions import MonitorKilledError
from typing import (
    Dict, 
    List, 
    Union,
    Any
)


WorkerMetrics = Dict[int, Dict[str, List[Union[int, float]]]] 


def handle_thread_loop_stop(
    loop: asyncio.AbstractEventLoop,
    monitor_name: str,
    running_monitors: Dict[str, bool]=None,

):
    running_monitors[monitor_name] = False
    loop.stop()
    loop.close()


def handle_monitor_stop(
    loop: asyncio.AbstractEventLoop=None,
    running_monitors: Dict[str, bool]=None,
    executor: ThreadPoolExecutor=None
):
    try:

        for monitor_name in running_monitors:
            running_monitors[monitor_name] = False

        executor.shutdown(wait=False, cancel_futures=True)

        if loop and loop.is_running():
            loop.stop()

        raise MonitorKilledError()
    
    except MonitorKilledError:
        executor.shutdown(wait=False, cancel_futures=True)
        pass

    except BrokenPipeError:
        executor.shutdown(wait=False, cancel_futures=True)
        raise MonitorKilledError()

    except RuntimeError:
        executor.shutdown(wait=False, cancel_futures=True)
        raise MonitorKilledError()
    
    finally:
        pass
    

class BaseMonitor:

    def __init__(self) -> None:
        self.active: Dict[str, List[int]] = defaultdict(list)
        self.collected: Dict[str, List[int]] = defaultdict(list)
        self.cpu_count = psutil.cpu_count()
        self.stage_metrics: Dict[str, List[Union[int, float]]] = {}
        self.visibility_filters: Dict[str, bool] = defaultdict(lambda: False)
        self.stage_type: Union[Any, None] = None
        self.worker_metrics: WorkerMetrics = defaultdict(dict)
        self.is_execute_stage = False

        self._background_monitors: Dict[str, asyncio.Task] = {}
        self._sync_background_monitors: Dict[str, asyncio.Future] = {}
        self._running_monitors: Dict[str, bool] = {}

        self._loop: Union[asyncio.AbstractEventLoop, None] = None
        self._executor: Union[ThreadPoolExecutor, None] = None

    def start_background_monitor_sync(
        self,
        monitor_name: str,
        interval_sec: Union[int, float]=1
    ):
        if self._executor is None:
            self._executor = ThreadPoolExecutor(
                max_workers=psutil.cpu_count(logical=False)
            )

            for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
                signal_type: signal = getattr(signal, signame)

                signal.signal(
                    signal_type, 
                    lambda sig, frame: handle_monitor_stop(  
                        running_monitors=self._running_monitors,
                        executor=self._executor
                    )
                )
        
        self._sync_background_monitors[monitor_name] = self._executor.submit(
            functools.partial(
                self._monitor_at_interval,
                monitor_name,
                interval_sec=interval_sec
            )
        )

    def aggregate_worker_stats(self):
        raise NotImplementedError('Aggregate worker stats method method must be implemented in a non-base Monitor class.')

    def _collect_worker_stats(self):

        monitor_stats: Dict[str, List[Union[int, float]]] = defaultdict(list)

        for worker_id in self.worker_metrics:
            for monitor_name, metrics in self.worker_metrics[worker_id].items():
                monitor_stats[monitor_name].append(metrics)

        return monitor_stats

    async def start_background_monitor(
        self,
        monitor_name: str,
        interval_sec: Union[int, float]=1
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._executor is None:
            self._executor = ThreadPoolExecutor(
                max_workers=psutil.cpu_count(logical=False)
            )

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            signal_type: signal = getattr(signal, signame)

            signal.signal(
                signal_type, 
                lambda sig, frame: handle_monitor_stop(  
                    loop=self._loop,
                    running_monitors=self._running_monitors,
                    executor=self._executor
                )
            )

        self._background_monitors[monitor_name] = self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._monitor_at_interval,
                monitor_name,
                interval_sec=interval_sec
            )
        )

    def update_monitor(str, monitor_name: str) -> Union[int, float]:
        raise NotImplementedError('Monitor background update method must be implemented in non-base Monitor class.')

    def store_monitor(self, monitor_name: str):
        self.collected[monitor_name] = list(self.active[monitor_name])
        del self.active[monitor_name]

    def trim_monitor_samples(
        self,
        monitor_name: str,
        trim_length: int
    ):
        if self.collected.get(monitor_name):
            self.collected[monitor_name][:trim_length]
        
    async def _update_background_monitor(
        self,
        monitor_name: str,
        interval_sec: Union[int, float]=1
    ):
        while self._running_monitors.get(monitor_name):
            await asyncio.sleep(interval_sec)
            self.update_monitor(monitor_name)

    def _monitor_at_interval(
        self, 
        monitor_name: str,
        interval_sec: Union[int, float]=1
):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        
        self._running_monitors[monitor_name] = True

        try:
            loop.run_until_complete(
                self._update_background_monitor(
                    monitor_name,
                    interval_sec=interval_sec
                )
            )

        except Exception:
            self._running_monitors[monitor_name] = False
            raise RuntimeError()
        
    def stop_background_monitor_sync(
        self,
        monitor_name: str
    ):

        self._running_monitors[monitor_name] = False
        self._sync_background_monitors[monitor_name].result()

        if self.active.get(monitor_name):
            self.collected[monitor_name].extend(
                list(self.active[monitor_name])
            )

    async def stop_background_monitor(
        self,
        monitor_name: str
    ):
        self._running_monitors[monitor_name] = False

        if not self._background_monitors[monitor_name].cancelled():
            await self._background_monitors[monitor_name]

        if self.active.get(monitor_name):
            self.collected[monitor_name].extend(
                list(self.active[monitor_name])
            )

    def stop_all_background_monitors_sync(self):

        for monitor_name in self._running_monitors.keys():
            self._running_monitors[monitor_name] = False

        for monitor in self._sync_background_monitors.values():
            monitor.result()

        for monitor_name in self._running_monitors.keys():
            self.collected[monitor_name] = list(self.active[monitor_name])

    async def stop_all_background_monitors(self):

        for monitor_name in self._running_monitors.keys():
            self._running_monitors[monitor_name] = False

        await asyncio.gather(list(self._background_monitors.values()))

        for monitor_name in self._running_monitors.keys():
            self.collected[monitor_name] = list(self.active[monitor_name])

    def close(self):
        if self._executor:
            self._executor.shutdown(wait=False, cancel_futures=True)


import asyncio
import time
import uuid
from typing import Dict, Coroutine, Any
from hedra.core.engines.types.common.base_engine import BaseEngine
from hedra.core.engines.types.common.ssl import get_default_ssl_context
from hedra.core.engines.types.common.timeouts import Timeouts
from .connection import UDPConnection
from .action import UDPAction
from .result import UDPResult
from .pool import Pool


class MercuryUDPClient(BaseEngine[UDPAction, UDPResult]):

    __slots__ = (
        'session_id',
        'timeouts',
        '_hosts',
        'closed',
        'sem',
        'pool',
        'active',
        'waiter',
        'ssl_context'
    )

    def __init__(self, concurrency: int=10**3, timeouts: Timeouts = Timeouts(), reset_connections: bool=False) -> None:
        super(
            MercuryUDPClient,
            self
        ).__init__()

        self.session_id = str(uuid.uuid4())
        self.timeouts = timeouts

        self.registered: Dict[str, UDPConnection] = {}
        self._hosts = {}
        self.closed = False

        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = Pool(concurrency, reset_connections=reset_connections)
        self.pool.create_pool()
        self.active = 0
        self.waiter = None

        self.ssl_context = get_default_ssl_context()

    def config_to_dict(self):
        return {
            'concurrency': self.pool.size,
            'timeouts': {
                'connect_timeout': self.timeouts.connect_timeout,
                'socket_read_timeout': self.timeouts.socket_read_timeout,
                'total_timeout': self.timeouts.total_timeout
            },
            'reset_connections': self.pool.reset_connections
        }
    
    async def set_pool(self, concurrency: int):
        self.sem = asyncio.Semaphore(value=concurrency)
        self.pool = Pool(concurrency, reset_connections=self.pool.reset_connections)
        self.pool.create_pool()

    async def prepare(self, action: UDPAction) -> Coroutine[Any, Any, None]:
        try:
            if action.url.is_ssl:
                action.ssl_context = self.ssl_context

            if self._hosts.get(action.url.hostname) is None:

                    socket_configs = await asyncio.wait_for(action.url.lookup(), timeout=self.timeouts.connect_timeout)
              
                    for ip_addr, configs in socket_configs.items():
                        for config in configs:

                            connection = UDPConnection()
                            
                            try:
                                await connection.make_connection(
                                    ip_addr,
                                    action.url.port,
                                    config,
                                    timeout=self.timeouts.connect_timeout
                                )

                                action.url.socket_config = config
                                action.url.ip_addr = ip_addr
                                action.url.has_ip_addr = True
                                break

                            except Exception as e: 
                                pass

                        if action.url.socket_config:
                            break

                    if action.url.socket_config is None:
                        raise Exception('Err. - No socket found.')
                    
                    self._hosts[action.url.hostname] = {
                        'ip_addr': action.url.ip_addr,
                        'socket_config': action.url.socket_config
                    }

            else:
                host_config = self._hosts[action.url.hostname]
                action.url.ip_addr = host_config.get('ip_addr')
                action.url.socket_config = host_config.get('socket_config')

            if action.is_setup is False:
                action.setup()

            self.registered[action.name] = action

        except Exception as e:   
            raise e

    def extend_pool(self, increased_capacity: int):
        self.pool.size += increased_capacity
        for _ in range(increased_capacity):
            self.pool.connections.append(
                UDPConnection(self.pool.reset_connections)
            )
        
        self.sem = asyncio.Semaphore(self.pool.size)

    def shrink_pool(self, decrease_capacity: int):
        self.pool.size -= decrease_capacity
        self.pool.connections = self.pool.connections[:self.pool.size]
        self.sem = asyncio.Semaphore(self.pool.size)

    async def execute_prepared_request(self, action: UDPAction) -> Coroutine[Any, Any, UDPResult]:
 
        response = UDPResult(action)
        response.wait_start = time.monotonic()
        self.active += 1
 
        async with self.sem:
            connection = self.pool.connections.pop()
            
            try:

                if action.hooks.listen:
                    event = asyncio.Event()
                    action.hooks.channel_events.append(event)
                    await event.wait()

                if action.hooks.before:
                    action = await self.execute_before(action)
                    action.setup()

                response.start = time.monotonic()

                await connection.make_connection(
                    action.url.ip_addr,
                    action.url.port,
                    action.url.socket_config,
                    timeout=self.timeouts.connect_timeout
                )

                response.connect_end = time.monotonic()
                
                if action.encoded_data:
                    if action.is_stream:
                        action.write_chunks(connection)

                    else:
                        connection.write(action.encoded_data)

                response.write_end = time.monotonic()
                
                if action.wait_for_response:
                    response.body = await connection.readuntil()
         
                response.complete = time.monotonic()

                self.pool.connections.append(connection)

                if action.hooks.after:
                    response = await self.execute_after(action, response)
                    action.setup()

                if action.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel.call(response, action.hooks.listeners)
                        ) for channel in action.hooks.channels
                    ])

                    for listener in action.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            listener.setup()
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()    

            except Exception as e:
                response.complete = time.monotonic()
                response.error = str(e)

                self.pool.connections.append(UDPConnection(reset_connection=self.pool.reset_connections))

            self.active -= 1
            if self.waiter and self.active <= self.pool.size:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return response

    async def close(self):
        if self.closed is False:
            await self.pool.close()
            self.closed = True



import asyncio
from typing import (
    List, 
    Dict, 
    Any, 
    Tuple
)
from collections import defaultdict
from hedra.core.graphs.stages.base.exceptions.process_killed_error import ProcessKilledError
from hedra.core.graphs.stages.base.parallel.stage_priority import StagePriority
from hedra.core.graphs.stages.base.parallel.batch_executor import BatchExecutor
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_types import StageTypes
from .transition import Transition
from .common.base_edge import BaseEdge


HistoryUpdate = Dict[Tuple[str, str], Any]


class TransitionGroup:

    def __init__(self) -> None:
        self.transitions: List[Transition]  = []
        self.transitions_by_type: Dict[StageTypes, List[Transition]] = defaultdict(list)
        self.destination_groups: Dict[Tuple[str, StageTypes], List[Transition]] = defaultdict(list)
        self.targets: Dict[str, Stage] = {}
        self.transitions_by_edge: Dict[Tuple[str, str,], Transition] = {}

        self.edges_by_name: Dict[str, BaseEdge] = {}
        self.adjacency_list: Dict[str, List[Transition]] = []
        self.transition_idx = 0
        self.cpu_pool_size = 0
        self._batched_transitions: List[List[Transition]] = []
        self._transition_configs: Dict[Tuple[str, str, str, int], int] = {}
        self._executors: List[BatchExecutor] = []

    @property
    def count(self):
        return len(self.transitions_by_edge)

    def __iter__(self):
        for transition in self.transitions_by_edge.values():
            yield transition

    def add_transition(self, transition: Transition):
        transition.edge.transition_idx = self.transition_idx
        self.destination_groups[(transition.from_stage.name, transition.to_stage.stage_type)].append(transition)
        self.transitions_by_type[transition.from_stage.stage_type].append(transition)
        self.transitions_by_edge[(transition.edge.source.name, transition.edge.destination.name)] = transition

        self.transition_idx += 1
    
    def sort_and_map_transitions(self):

        for edge_key, transition in self.transitions_by_edge.items():

            transition.edges = [
                group_transition.edge for group_transition in self.transitions_by_edge.values()
            ]

            destinations = self.adjacency_list[transition.edge.source.name]
            transition.destinations = [
                transition.edge.destination.name
            ]

            if len(destinations)> 1:
                transition.edge.setup()
                transition.edge.split([transition.edge for transition in destinations])

                transition.destinations = [
                    destination_transition.edge.destination.name for destination_transition in destinations
                ]

            self.transitions_by_edge[edge_key] = transition

        executor = BatchExecutor(max_workers=self.cpu_pool_size)
        batched_transitions = executor.partion_prioritized_stage_batches(
            list(self.transitions_by_edge.values())
        )

        prioritized_groups: List[Tuple[str, str, str, int]] = []
        for group in batched_transitions:
            
            group_priorities: List[int] = []
            for transition_config in group:
                source, destination, _, workers = transition_config
                
                transition = self.transitions_by_edge.get((
                    source,
                    destination
                ))

                transition_priority: StagePriority = transition.edge.source.priority_level

                if transition_priority == StagePriority.AUTO:
                    group_priorities.append(0)

                else:
                    group_priorities.append(transition_priority.value)

            group_priority = 0
            if len(group_priorities) > 0:
                group_priority = max(group_priorities)
                prioritized_groups.append((
                    group_priority,
                    group
                ))

        sorted_batches = list(sorted(
            prioritized_groups,
            key=lambda prioritized_group: prioritized_group[0],
            reverse=True
        ))

        self._batched_transitions = [
            group for _, group in sorted_batches
        ]

        for group in self._batched_transitions:
            for source, destination, _, workers in group:
                self._transition_configs[(source, destination)] = workers

        executor.close()

    async def execute_group(self):
        results: List[Any] = []

        for group in self._batched_transitions:
            sorted_group = sorted(
                group,
                key=lambda transition: transition[3],
                reverse=True
            )
                        
            group_results = await asyncio.gather(*[
                asyncio.create_task(
                    self._execute_transition(transition_config)
                ) for transition_config in sorted_group
            ])

            results.extend(group_results)

        return results
    
    async def _execute_transition(self, transition_config: Tuple[str, str, str, int]) -> Any:

        source, destination, _, workers = transition_config

        transition = self.transitions_by_edge.get((
            source,
            destination
        ))
    
        if workers > 0:
            transition.edge.source.workers = workers
            executor = BatchExecutor(max_workers=workers)
            transition.edge.source.executor = executor
            self._executors.append(executor)
        
        return await transition.execute()


import json
from typing import Any, Dict, Union
from hedra.core.engines.types.playwright import PlaywrightResult
from .base_processed_result import BaseProcessedResult


class PlaywrightProcessedResult(BaseProcessedResult):

    __slots__ = (
        'event_id',
        'action_id',
        'url',
        'headers',
        'command',
        'selector',
        'x_coord',
        'y_coord',
        'frame',
        'timings'
    )

    def __init__(
        self, 
        stage: str, 
        result: PlaywrightResult
    ) -> None:
        super(
            PlaywrightProcessedResult,
            self
        ).__init__(
            stage,
            result
        )

        self.url = result.url
        self.headers = result.headers
        self.command = result.command
        self.selector = result.selector
        self.x_coord = result.x_coord
        self.y_coord = result.y_coord
        self.frame = result.frame

        self.time = result.complete - result.start

        self.timings = {
            'total': self.time,
            'waiting': result.start - result.wait_start,
            'connecting': result.connect_end - result.start,
            'writing': result.write_end - result.connect_end,
            'reading': result.complete - result.write_end
        }

    def to_dict(self) -> Dict[str, Union[str, int, float]]:

        return {
            'name': self.name,
            'stage': self.stage,
            'shortname': self.shortname,
            'checks': [check.__name__ for check in self.checks],
            'error': str(self.error),
            'time': self.time,
            'type': self.type,
            'source': self.source,
            'url': self.url,
            'headers': self.headers,
            'command': self.command,
            'selector': self.selector,
            'x_coord': self.x_coord,
            'y_coord': self.y_coord,
            'frame': self.frame,
            **self.timings
        }

    def serialize(self) -> str:
        return json.dumps(
            self.to_dict()
        )


# # This is an ugly patch for: https://github.com/aio-libs/aiopg/issues/837
# import selectors  # isort:skip # noqa: F401

# selectors._PollLikeSelector.modify = (  # type: ignore
#     selectors._BaseSelectorImpl.modify  # type: ignore
# )

import uuid
from typing import List, Dict
from hedra.logging import HedraLogger
from hedra.reporting.experiment.experiments_collection import ExperimentMetricsCollectionSet
from hedra.reporting.processed_result.types.base_processed_result import BaseProcessedResult
from hedra.reporting.metric.stage_streams_set import StageStreamsSet
from hedra.reporting.metric import (
    MetricsSet,
    MetricType
)
from hedra.reporting.system.system_metrics_set import (
    SystemMetricsSet,
    SessionMetricsCollection,
    SystemMetricsCollection
)
from .postgres_config import PostgresConfig

try:
    import sqlalchemy
    from sqlalchemy.dialects.postgresql import UUID
    from sqlalchemy.schema import CreateTable
    from sqlalchemy.ext.asyncio import create_async_engine
    from sqlalchemy.ext.asyncio.engine import (
        AsyncEngine, 
        AsyncConnection,
        AsyncTransaction
    )
    
    has_connector = True

except Exception:
    UUID = None
    sqlalchemy = None
    create_engine = None
    has_connector = False


class Postgres:

    def __init__(self, config: PostgresConfig) -> None:
        self.host = config.host
        self.database = config.database
        self.username = config.username
        self.password = config.password

        self.events_table_name = config.events_table
        self.metrics_table_name = config.metrics_table
        self.streams_table_name = config.streams_table

        self.experiments_table_name = config.experiments_table
        self.variants_table_name = f'{config.experiments_table}_variants'
        self.mutations_table_name = f'{config.experiments_table}_mutations'

        self.shared_metrics_table_name = f'{config.metrics_table}_shared'
        self.errors_table_name = f'{config.metrics_table}_errors'
        self.custom_metrics_table_name = f'{config.metrics_table}_custom'

        self.session_system_metrics_table_name = f'{config.system_metrics_table}_session'
        self.stage_system_metrics_table_name = f'{config.system_metrics_table}_stage'
        
        self._engine = None
        self.metadata = sqlalchemy.MetaData()

        self._events_table = None
        self._metrics_table = None
        self._streams_table = None

        self._experiments_table = None
        self._variants_table = None
        self._mutations_table = None

        self._shared_metrics_table = None
        self._errors_table = None
        self._custom_metrics_table = None

        self._session_system_metrics_table = None
        self._stage_system_metrics_table = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()
        self.sql_type = 'Postgresql'

        self.metric_types_map = {
            MetricType.COUNT: lambda field_name: sqlalchemy.Column(
                field_name, 
                sqlalchemy.BIGINT
            ),
            MetricType.DISTRIBUTION: lambda field_name: sqlalchemy.Column(
                field_name, 
                sqlalchemy.FLOAT
            ),
            MetricType.SAMPLE: lambda field_name: sqlalchemy.Column(
                field_name, 
                sqlalchemy.FLOAT
            ),
            MetricType.RATE: lambda field_name: sqlalchemy.Column(
                field_name, 
                sqlalchemy.FLOAT
            ),
        }

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to {self.sql_type} instance at - {self.host} - Database: {self.database}')
        
        connection_uri = 'postgresql+asyncpg://'

        if self.username and self.password:
            connection_uri = f'{connection_uri}{self.username}:{self.password}@'

        self._engine: AsyncEngine = await create_async_engine(
            f'{connection_uri}{self.host}/{self.database}',
            echo=False
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to {self.sql_type} instance at - {self.host} - Database: {self.database}')

    async def submit_session_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Session System Metrics to Table - {self.session_system_metrics_table_name}')

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics to Table - {self.session_system_metrics_table_name} - Initiating transaction')
                
                if self._session_system_metrics_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Session System Metrics table - {self.session_system_metrics_table_name} - if not exists')

                    session_system_metrics_table = sqlalchemy.Table(
                        self.session_system_metrics_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                        sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('group', sqlalchemy.TEXT()),
                        sqlalchemy.Column('median', sqlalchemy.FLOAT),
                        sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                        sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                        sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                        sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                        sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
                    )

                    for quantile in SystemMetricsSet.quantiles:
                        session_system_metrics_table.append_column(
                            sqlalchemy.Column(f'quantile_{quantile}th', sqlalchemy.FLOAT)
                        )

                    await connection.execute(
                        CreateTable(
                            session_system_metrics_table, 
                            if_not_exists=True
                        )
                    )
                
                    self._session_system_metrics_table = session_system_metrics_table

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Session System Metrics table - {self.session_system_metrics_table_name}')

                rows: List[SessionMetricsCollection] = []
                
                for metrics_set in system_metrics_sets:
                    for monitor_metrics in metrics_set.session_cpu_metrics.values():
                        rows.append(monitor_metrics)
                        
                    for  monitor_metrics in metrics_set.session_memory_metrics.values():
                        rows.append(monitor_metrics)
            
                for metrics_set in rows:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics - {metrics_set.name}:{metrics_set.group}')

                    await connection.execute(
                        self._streams_table.insert(values=metrics_set.record)
                    )
                    
                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Session System Metrics to Table - {self.session_system_metrics_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Session System Metrics to Table - {self.session_system_metrics_table_name}')

    async def submit_stage_system_metrics(self, system_metrics_sets: List[SystemMetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Stage System Metrics to Table - {self.stage_system_metrics_table_name}')

        
        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics to Table - {self.stage_system_metrics_table_name} - Initiating transaction')
                if self._stage_system_metrics_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Stage System Metrics table - {self.stage_system_metrics_table_name} - if not exists')

                    stage_system_metrics_table = sqlalchemy.Table(
                        self.stage_system_metrics_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                        sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('group', sqlalchemy.TEXT()),
                        sqlalchemy.Column('median', sqlalchemy.FLOAT),
                        sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                        sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                        sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                        sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                        sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
                    )

                    for quantile in SystemMetricsSet.quantiles:
                        stage_system_metrics_table.append_column(
                            sqlalchemy.Column(f'quantile_{quantile}th', sqlalchemy.FLOAT)
                        )

                    await connection.execute(
                        CreateTable(
                            stage_system_metrics_table, 
                            if_not_exists=True
                        )
                    )
                
                    self._stage_system_metrics_table = stage_system_metrics_table

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Stage System Metrics table - {self.stage_system_metrics_table_name}')

                rows: List[SystemMetricsCollection] = []
                
                for metrics_set in system_metrics_sets:
                    cpu_metrics = metrics_set.cpu
                    memory_metrics = metrics_set.memory

                    for stage_name, stage_cpu_metrics in  cpu_metrics.metrics.items():

                        for monitor_metrics in stage_cpu_metrics.values():
                            rows.append(monitor_metrics)

                        stage_memory_metrics = memory_metrics.metrics.get(stage_name)
                        for monitor_metrics in stage_memory_metrics.values():
                            rows.append(monitor_metrics)

                        stage_mb_per_vu_metrics = metrics_set.mb_per_vu.get(stage_name)
                        
                        if stage_mb_per_vu_metrics:
                            rows.append(stage_mb_per_vu_metrics)
            
                for metrics_set in rows:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics - {metrics_set.name}:{metrics_set.group}')

                    await connection.execute(
                        self._streams_table.insert(values=metrics_set.record)
                    )
                    
                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Stage System Metrics to Table - {self.stage_system_metrics_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Stage System Metrics to Table - {self.stage_system_metrics_table_name}')

    async def submit_streams(self, stream_metrics: Dict[str, StageStreamsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Streams to Table - {self.streams_table_name}')

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams to Table - {self.streams_table_name} - Initiating transaction')
            
                for stage_name, stream in stream_metrics.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams - {stage_name}:{stream.stream_set_id}')

                    if self._streams_table is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Streams table - {self.streams_table_name} - if not exists')

                        stream_table = sqlalchemy.Table(
                            self.streams_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('group', sqlalchemy.TEXT()),
                            sqlalchemy.Column('median', sqlalchemy.FLOAT),
                            sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                            sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                            sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                            sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                            sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
                        )

                        for quantile in stream.quantiles:
                            stream_table.append_column(
                                sqlalchemy.Column(f'quantile_{quantile}th', sqlalchemy.FLOAT)
                            )

                        await connection.execute(
                            CreateTable(
                                stream_table, 
                                if_not_exists=True
                            )
                        )
                    
                        self._streams_table = stream_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Streams table - {self.streams_table_name}')

                    for group_name, group in stream.grouped.items():
                        await connection.execute(
                            self._streams_table.insert(values={
                                'name': f'{stage_name}_streams',
                                'stage': stage_name,
                                'group': group_name,
                                **group
                            })
                        )
                    
                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Streams to Table - {self.streams_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Streams to Table - {self.streams_table_name}')

    async def submit_experiments(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Experiments to Table - {self.experiments_table_name}')
        
        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiments to Table - {self.experiments_table_name} - Initiating transaction')
            
                for experiment in experiments_metrics.experiment_summaries:

                    if self._experiments_table is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Experiments table - {self.experiments_table_name} - if not exists')

                        experiments_table = sqlalchemy.Table(
                            self.experiments_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                            sqlalchemy.Column('experiment_name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('experiment_randomized', sqlalchemy.Boolean),
                            sqlalchemy.Column('experiment_completed', sqlalchemy.BIGINT),
                            sqlalchemy.Column('experiment_succeeded', sqlalchemy.BIGINT),
                            sqlalchemy.Column('experiment_failed', sqlalchemy.BIGINT),
                            sqlalchemy.Column('experiment_median_aps', sqlalchemy.FLOAT),
                        )

                        await connection.execute(CreateTable(experiments_table, if_not_exists=True))

                        self._experiments_table = experiments_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Experiments table - {self.experiments_table_name}')
                    
                    await connection.execute(self._experiments_table.insert(
                        values=experiment.record
                    ))
                        
                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Experiments to Table - {self.experiments_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Experiments to Table - {self.experiments_table_name}')

    async def submit_variants(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Variants to Table - {self.variants_table_name}')
        
        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction
            
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variants to Table - {self.variants_table_name} - Initiating transaction')
            
                for variant in experiments_metrics.variant_summaries:

                    if self._variants_table is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Variants table - {self.variants_table_name} - if not exists')

                        variants_table = sqlalchemy.Table(
                            self.variants_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                            sqlalchemy.Column('variant_name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('variant_experiment', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('variant_weight', sqlalchemy.FLOAT),
                            sqlalchemy.Column('variant_distribution', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('variant_distribution_interval', sqlalchemy.FLOAT),
                            sqlalchemy.Column('variant_completed', sqlalchemy.BIGINT),
                            sqlalchemy.Column('variant_succeeded', sqlalchemy.BIGINT),
                            sqlalchemy.Column('variant_failed', sqlalchemy.BIGINT),
                            sqlalchemy.Column('variant_actions_per_second', sqlalchemy.FLOAT),
                            sqlalchemy.Column('variant_ratio_completed', sqlalchemy.FLOAT),
                            sqlalchemy.Column('variant_ratio_succeeded', sqlalchemy.FLOAT),
                            sqlalchemy.Column('variant_ratio_failed', sqlalchemy.FLOAT),
                            sqlalchemy.Column('variant_ratio_aps', sqlalchemy.FLOAT),
                        )

                        await connection.execute(CreateTable(variants_table, if_not_exists=True))

                        self._variants_table = variants_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Variants table - {self.variants_table_name}')
                    
                    await connection.execute(self._variants_table.insert(
                        values=variant.record
                    ))
                        
                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Variants to Table - {self.variants_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Variants to Table - {self.variants_table_name}')

    async def submit_mutations(self, experiments_metrics: ExperimentMetricsCollectionSet):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Mutations to Table - {self.mutations_table_name}')
        
        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutations to Table - {self.mutations_table_name} - Initiating transaction')
            
                for mutation in experiments_metrics.mutation_summaries:

                    if self._mutations_table is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Mutations table - {self.mutations_table_name} - if not exists')

                        mutations_table = sqlalchemy.Table(
                            self.mutations_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', sqlalchemy.Integer, primary_key=True),
                            sqlalchemy.Column('mutation_name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('mutation_experiment_name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('mutation_variant_name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('mutation_chance', sqlalchemy.FLOAT),
                            sqlalchemy.Column('mutation_targets', sqlalchemy.VARCHAR(8192)),
                            sqlalchemy.Column('mutation_type', sqlalchemy.VARCHAR(255)),
                        )

                        await connection.execute(CreateTable(mutations_table, if_not_exists=True))

                        self._mutations_table = mutations_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Mutations table - {self.mutations_table_name}')
                    
                    await connection.execute(self._mutations_table.insert(
                        values=mutation.record
                    ))
                        
                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Mutations to Table - {self.mutations_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Mutations to Table - {self.mutations_table_name}')

    async def submit_events(self, events: List[BaseProcessedResult]):
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name}')

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name} - Initiating transaction')

                for event in events:

                    if self._events_table is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Events table - {self.events_table_name} - if not exists')

                        events_table = sqlalchemy.Table(
                            self.events_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', UUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
                            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('time', sqlalchemy.FLOAT),
                            sqlalchemy.Column('succeeded', sqlalchemy.Boolean),
                        )
                        
                        await connection.execute(CreateTable(events_table, if_not_exists=True))
                        self._events_table = events_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Events table - {self.events_table_name}')
                    
                    await connection.execute(self._events_table.insert(
                        values=event.record
                    ))
                
                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Events to Table - {self.events_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to Table - {self.events_table_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name}')

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name} - Initiating transaction')
                
                if self._shared_metrics_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Shared Metrics table - {self.shared_metrics_table_name} - if not exists')

                    stage_metrics_table = sqlalchemy.Table(
                        self.shared_metrics_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', UUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
                        sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('total', sqlalchemy.BIGINT),
                        sqlalchemy.Column('succeeded', sqlalchemy.BIGINT),
                        sqlalchemy.Column('failed', sqlalchemy.BIGINT),
                        sqlalchemy.Column('actions_per_second', sqlalchemy.FLOAT)
                    )

                    await connection.execute(CreateTable(stage_metrics_table, if_not_exists=True))
                    self._shared_metrics_table = stage_metrics_table

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Shared Metrics table - {self.shared_metrics_table_name}')

                for metrics_set in metrics_sets:
                    await connection.execute(
                        self._metrics_table.insert(values={
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': 'common',
                            **metrics_set.common_stats
                        })
                    )

                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics to Table - {self.shared_metrics_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to Table - {self.shared_metrics_table_name}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name}')

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name} - Initiating transaction')

                for metrics_set in metrics:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                    if self._metrics_table is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Metrics table - {self.metrics_table_name} - if not exists')

                        metrics_table = sqlalchemy.Table(
                            self.metrics_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', UUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
                            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('median', sqlalchemy.FLOAT),
                            sqlalchemy.Column('mean', sqlalchemy.FLOAT),
                            sqlalchemy.Column('variance', sqlalchemy.FLOAT),
                            sqlalchemy.Column('stdev', sqlalchemy.FLOAT),
                            sqlalchemy.Column('minimum', sqlalchemy.FLOAT),
                            sqlalchemy.Column('maximum', sqlalchemy.FLOAT)
                        )

                        for quantile in metrics_set.quantiles:
                            metrics_table.append_column(
                                sqlalchemy.Column(f'{quantile}', sqlalchemy.FLOAT)
                            )

                        for custom_field_name, sql_alchemy_type in metrics_set.custom_schemas:
                            metrics_table.append_column(custom_field_name, sql_alchemy_type) 

                        await connection.execute(CreateTable(metrics_table, if_not_exists=True))
                        self._metrics_table = metrics_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Metrics table - {self.metrics_table_name}')

                    for group_name, group in metrics_set.groups.items():
                        await connection.execute(
                            self._metrics_table.insert(values={
                                **group.record,
                                'group': group_name
                            })
                        )

                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics to Table - {self.metrics_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to Table - {self.metrics_table_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics - Initiating transaction')

                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Custom Metrics to table - {self.custom_metrics_table_name}')

                if self._custom_metrics_table is None:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Custom Metrics table - {self.custom_metrics_table_name} - if not exists')

                    custom_metrics_table = sqlalchemy.Table(
                        self.custom_metrics_table_name,
                        self.metadata,
                        sqlalchemy.Column('id', UUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
                        sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                        sqlalchemy.Column('group', sqlalchemy.VARCHAR(255)),
                    )

                    for metrics_set in metrics_sets:
                        for custom_metric_name, custom_metric in metrics_set.custom_metrics.items():

                            custom_metrics_table.append_column(
                                self.metric_types_map.get(
                                    custom_metric.metric_type,
                                    lambda field_name: sqlalchemy.Column(
                                        field_name, 
                                        sqlalchemy.FLOAT
                                    )
                                )(custom_metric_name)
                            )

                    await connection.execute(
                        CreateTable(custom_metrics_table, if_not_exists=True)
                    )

                    self._custom_metrics_table = custom_metrics_table

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Custom Metrics table - {self.custom_metrics_table_name}')

                
                for metrics_set in metrics_sets:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
                        
                    await connection.execute(
                        self._custom_metrics_table.insert(values={
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': 'custom',
                            **{
                                custom_metric_name: custom_metric.metric_value for custom_metric_name, custom_metric in metrics_set.custom_metrics.items()
                            }
                        })
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Custom Metrics to table - {self.custom_metrics_table_name}')

                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics - Transaction committed')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name}')

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection

            async with connection.begin() as transaction:
                transaction: AsyncTransaction = transaction

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name} - Initiating transaction')

                for metrics_set in metrics_sets:
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                    if self._errors_table is None:
                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Creating Error Metrics table - {self.errors_table_name} - if not exists')

                        errors_table = sqlalchemy.Table(
                            self.errors_table_name,
                            self.metadata,
                            sqlalchemy.Column('id', UUID(as_uuid=True), primary_key=True, default=uuid.uuid4),
                            sqlalchemy.Column('name', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('stage', sqlalchemy.VARCHAR(255)),
                            sqlalchemy.Column('error_message', sqlalchemy.TEXT),
                            sqlalchemy.Column('error_count', sqlalchemy.Integer)
                        )  


                        await connection.execute(CreateTable(errors_table, if_not_exists=True))
                        self._errors_table = errors_table

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Created or set Error Metrics table - {self.errors_table_name}')

                    for error in metrics_set.errors:
                        await connection.execute(
                            self._errors_table.insert(values={
                                'name': metrics_set.name,
                                'stage': metrics_set.stage,
                                'error_message': error.get('message'),
                                'error_count': error.get('count')
                            })
                        )

                await transaction.commit()
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics to Table - {self.errors_table_name} - Transaction committed')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to Table - {self.errors_table_name}')
        
    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Session Closed - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connectiion to {self.sql_type} at - {self.host}')



    



import os
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class XMLConfig(BaseModel):
    events_filepath: str=os.path.join(
        os.getcwd(),
        'events.xml'
    )
    metrics_filepath: str=os.path.join(
        os.getcwd(),
        'metrics.xml'
    )
    experiments_filepath: str=os.path.join(
        os.getcwd(),
        'experiments.xml'
    )
    streams_filepath: str=os.path.join(
        os.getcwd(),
        'streams.xml'
    )
    system_metrics_filepath: str=os.path.join(
        os.getcwd(),
        'system_metrics.xml'
    )
    overwrite: bool=True
    reporter_type: ReporterTypes=ReporterTypes.XML



import uuid
import asyncio
import json
from typing import (
    List, 
    Dict, 
    Any,
    Union,
    Coroutine
)
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .kafka_connector_config import KafkaConnectorConfig


try:

    from aiokafka import AIOKafkaConsumer
    from aiokafka.structs import ConsumerRecord
    has_connector = True

except Exception:
    AIOKafkaConsumer = object
    ConsumerRecord = object
    has_connector = False


class KafkaConnector:
    connector_type=ConnectorType.Kafka

    def __init__(
        self, 
        config: KafkaConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        
        self.host = config.host
        self.client_id = config.client_id
        self.stage = stage
        self.parser_config = parser_config

        self.topic = config.topic
        self.partition = config.partition

        self.compression_type = config.compression_type
        self.timeout = config.timeout
        self.enable_idempotence = config.idempotent or True
        self.options: Dict[str, Any] = config.options or {}
        self._consumer = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.loop: Union[asyncio.AbstractEventLoop, None] = None

        self.logger = HedraLogger()
        self.logger.initialize()
        self.parser = Parser()

    async def connect(self):

        self.loop = asyncio.get_event_loop()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to Kafka at - {self.host}')

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Compression Type: {self.compression_type}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Connection Timeout: {self.timeout}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - Idempotent: {self.enable_idempotence}')

        for option_name, option in self.options.items():
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Using Kafka Options - {option_name.capitalize()}: {option}')


        self._consumer = AIOKafkaConsumer(
            self.topic,
            loop=self.loop,
            bootstrap_servers=self.host,
            client_id=self.client_id,
            compression_type=self.compression_type,
            request_timeout_ms=self.timeout,
            enable_idempotence=self.enable_idempotence,
            **self.options
        )

        await self._consumer.start()

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to Kafka at - {self.host}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> ExecuteStageSummaryValidator:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        
        actions = await self.load_data(
            options=options
        )

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])

    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data(
            options=options
        )

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        data: Dict[str, List[ConsumerRecord]] = await self._consumer.getmany(
            timeout_ms=self.timeout,
            max_records=options.get('max_records')
        )

        records: List[Dict[str, Any]] = []

        for messages in data.values():
            for message in messages:
                records.append(
                    json.loads(message.value)
                )

        return records


    async def close(self):
        await self._consumer.stop()



from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    GraphiteConfig
)


class SubmitGraphiteResultsStage(Submit):
    config=GraphiteConfig(
        host='localhost',
        port=2003
    )


from .reporter_connect import reporter_connect
from .reporter_close import reporter_close
from .process_events import process_events
from .process_shared import process_shared
from .process_metrics import process_metrics
from .process_custom import process_custom
from .process_errors import process_errors


from http.cookies import SimpleCookie
from pydantic import (
    BaseModel
)
from typing import (
    Dict, 
    Union
)


class Response:
    def __init__(
        self,
        path: str,
        method: str,
        headers: Dict[str, str]={},
        data: Union[BaseModel, str, None]=None
    ):
        self.path = path
        self.method = method
        self.headers = headers
        self.data = data
        self._cookies: Union[
            Dict[str, str],
            None
        ] = None

    @property
    def cookies(self):

        if self._cookies is None:

            cookies = self.headers.get('cookie')
            self._cookies = {}

            if cookies:

                parsed_cookies = SimpleCookie()
                parsed_cookies.load(cookies)

                self._cookies =  {
                    name: morsel.value for name, morsel in parsed_cookies.items()
                }

        return self._cookies



from enum import Enum

class GraphStatus(Enum):
    IDLE='IDLE'
    INITIALIZING='INITIALIZING'
    VALIDATING='VALIDATING'
    ASSEMBLING='ASSEMBLING'
    RUNNING='RUNNING'
    COMPLETE='COMPLETE'
    CANCELLED='CANCELLED'
    FAILED='FAILED'


import binascii
import collections
from typing import Optional


def raw_data_repr(data: Optional[bytes]) -> str:
    if not data:
        return "None"
    r = binascii.hexlify(data).decode('ascii')
    if len(r) > 20:
        r = r[:20] + "..."
    return "<hex:" + r + ">"


HeaderValidationFlags = collections.namedtuple(
    'HeaderValidationFlags',
    ['is_client', 'is_trailer', 'is_response_header', 'is_push_promise']
)



import asyncio
import math
import os
import psutil
import statistics
from hedra.distributed.models.http import Limit
from typing import Union, List
from .base_limiter import BaseLimiter


class CPUAdaptiveLimiter(BaseLimiter):

    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop",
        "_last_check",
        "_cpu_limit",
        "_current_time",
        "_previous_count",
        "_current_cpu",
        "_max_queue",
        "_sample_task",
        "_running",
        "_process",
        "_max_fast_backoff",
        "_min_backoff",
        "_history"
    )

    def __init__(
        self,
        limit: Limit
     ) -> None:
        super().__init__(
            limit.max_requests,
            limit.period,
            reject_requests=limit.reject_requests
        )

        cpu_limit = limit.cpu_limit
        if cpu_limit is None:
            cpu_limit = 50

        self._cpu_limit = cpu_limit
        self._backoff = limit.backoff
        self._min_backoff = self._backoff
        self._max_fast_backoff = math.ceil(self._backoff * 10)
        self._max_backoff = math.ceil(self._max_fast_backoff * 10)
        self._last_check = self._loop.time()
        self._current_time = self._loop.time()
        self._previous_count = limit.max_requests

        self._history: List[float] = []

        self._max_queue = limit.max_requests
        self._sample_task: Union[asyncio.Task, None] = None
        self._running = False
        self._activate_limit = False
        self._process = psutil.Process(os.getpid())

        self._current_cpu = self._process.cpu_percent()
        self._history.append(self._current_cpu)

    def has_capacity(self, amount: float = 1) -> bool:

        elapsed = self._loop.time() - self._last_check

        self._backoff = max(
            self._backoff - (1 / self._rate_per_sec * elapsed), 
            self._min_backoff
        )

        if (self._loop.time() - self._current_time) > self.time_period:
            self._current_time = math.floor(
                self._loop.time()/self.time_period
            ) * self.time_period

            self._previous_count = self._level
            self._level = 0

        self._rate_per_sec = (
            self._previous_count * (
                self.time_period - (self._loop.time() - self._current_time)
            )/self.time_period
        ) + (self._level + amount)

        if self._rate_per_sec < self.max_rate:

            for fut in self._waiters.values():
                if not fut.done():
                    fut.set_result(True)
                    break
        
        self._last_check = self._loop.time()

        return self._rate_per_sec <= self.max_rate
    
    async def acquire(
        self, 
        amount: float = 1,
    ):
        
        if not self._running:
            self._running = True
            self._sample_task = asyncio.create_task(
                self._sample_cpu()
            )
        
        if amount > self.max_rate:
            raise ValueError("Can't acquire more than the maximum capacity")


        task = asyncio.current_task(
            loop=self._loop
        )

        assert task is not None

        rejected = False

        while not self.has_capacity(amount) or self._activate_limit:

            fut = self._loop.create_future()
            try:

                self._waiters[task] = fut

                await asyncio.wait_for(
                    asyncio.shield(fut), 
                    timeout=self._backoff
                )

                if self._activate_limit:
                    await asyncio.sleep(self._backoff)
                    self._max_fast_backoff = min(
                        self._max_fast_backoff + (1/math.sqrt(self._rate_per_sec)),
                        self._max_backoff
                    )

            except asyncio.TimeoutError:
                pass
                
            fut.cancel()
            
            rejected = True

        self._backoff = min(self._backoff * 2, self._max_fast_backoff)
        self._waiters.pop(task, None)
        self._level += amount

        return rejected
    
    async def _sample_cpu(self):
        while  self._running:

            self._current_cpu = self._process.cpu_percent()
            self._history.append(self._current_cpu)

            elapsed = self._loop.time() - self._last_check

            if elapsed > self.time_period:
                self._history.pop(0)

            if self._current_cpu >= self._cpu_limit:
                self._activate_limit = True

            elif statistics.median(self._history) < self._cpu_limit:
                self._activate_limit = False
                self._max_fast_backoff = max(
                    self._max_fast_backoff - (1/self._rate_per_sec),
                    self._min_backoff
                )

            await asyncio.sleep(0.1)

    async def close(self):
        self._running = False

        self._sample_task.cancel()
        if not self._sample_task.cancelled():
            try:

                await self._sample_task

            except (
                asyncio.CancelledError,
                asyncio.InvalidStateError
            ):
                pass


import asyncio
import contextvars
import functools


async def awaitable(func, *args, **kwargs):
    """
    NOTE: Due to a Python 3.9 bug with OSX
          we can't use asyncio.to_thread(), so we
          have to re-implement it here.
    """
    loop = asyncio.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)







import asyncio
import uuid
import functools
import signal
import psutil
from concurrent.futures import ThreadPoolExecutor
from typing import (
    List, 
    Dict, 
    Any,
    Coroutine
)
from hedra.core.engines.client.config import Config
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.data.parsers.parser import Parser
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.logging import HedraLogger
from .bigtable_connector_config import BigTableConnectorConfig

try:
    from google.cloud import bigtable
    from google.auth import load_credentials_from_file
    from google.cloud.bigtable.row_data import PartialRowsData
    has_connector = True

except Exception:
    bigtable = None
    Credentials = None
    has_connector = False


def handle_loop_stop(
    signame, 
    executor: ThreadPoolExecutor, 
    loop: asyncio.AbstractEventLoop
): 
    try:
        executor.shutdown(wait=False, cancel_futures=True) 
        loop.stop()
    except Exception:
        pass


class BigTableConnector:
    connector_type = ConnectorType.BigTable

    def __init__(
        self, 
        config: BigTableConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:

        self.service_account_json_path = config.service_account_json_path
        self.instance_id = config.instance_id
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))

        self.instance = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.stage = stage
        self.parser_config = parser_config

        self.logger = HedraLogger()
        self.logger.initialize()

        self._table_name = config.table_name
        self._column_family_name = None
        self._table = None  
        self._columns = None
    
        self.credentials = None
        self.client = None
        self._loop = asyncio.get_event_loop()

        self.parser = Parser()

    async def connect(self):

        for signame in ('SIGINT', 'SIGTERM', 'SIG_IGN'):
            self._loop.add_signal_handler(
                getattr(signal, signame),
                lambda signame=signame: handle_loop_stop(
                    signame,
                    self._executor,
                    self._loop
                )
            )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opening amd authorizing connection to Google Cloud - Loading account config from - {self.service_account_json_path}')
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Opening session - {self.session_uuid}')

        credentials, project_id = load_credentials_from_file(self.service_account_json_path)
        self.client = bigtable.Client(
            project=project_id,
            credentials=credentials,
            admin=True
        )
        self.instance = self.client.instance(self.instance_id)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Created Client Instance - ID:{self.instance_id}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Opened connection to Google Cloud - Loaded account config from - {self.service_account_json_path}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        actions = await self.load_data()

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data()

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        self._table = self.instance.table(self._table_name)

        data_rows: PartialRowsData = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._table.read_rows
            )
        )

        return [
            row.to_dict() for row in data_rows
        ]
    
    async def close(self):
        self._executor.shutdown(cancel_futures=True)



from .alpha import AlphaDistribution
from .anglit import AnglitDistribution
from .arcsine import ArcsineDistribution
from .argus import ArgusDistribution
from .beta_prime import BetaPrimeDistribution
from .beta import BetaDistribution
from .bradford import BradfordDistribution
from .burr_12 import Burr12Distribution
from .burr import BurrDistribution
from .cauchy import CauchyDistribution
from .chi_squared import ChiSquaredDistribution
from .chi import ChiDistribution
from .cosine import CosineDistribution
from .crystal_ball import CrystalDistribution
from .dgamma import DGammaDistribution
from .dweibull import DWeibullDistribution
from .erlang import ErlangDistribution
from .exponential_normal import ExponentialNormalDistribution
from .exponential_power import ExponentialPowerDistribution
from .exponential import ExponentialDistribution
from .f_distribution import FDistribution
from .fatigue_life import FatigueLifeDistribution
from .fisk import FiskDistribution
from .folded_cauchy import FoldedCauchyDistribution
from .folded_normal import FoldedNormalDistribution
from .gamma import GammaDistribution
from .gauss_hypergeometric import GaussHypergeometricDistribution
from .generalized_exponential import GeneralizedExponentialDistribution
from .generalized_extreme import GeneralizedExtremeDistribution
from .generalized_gamma import GeneralizedGammaDistribution
from .generalized_half_logistic import GeneralizedHalfLogisticDistribution
from .generalized_hyperbolic import GeneralizedHyperbolicDistribution
from .generalized_inverse_gauss import GeneralizedInverseGaussDistribution
from .generalized_logistic import GeneralizedLogisticDistribution
from .generalized_normal import GeneralizedNormalDistribution
from .generalized_pareto import GeneralizedParetoDistribution
from .gibrat import GibratDistribution
from .gompertz import GompertzDistribution
from .gumbel_l import GumbelLDistribution
from .gumbel_r import GumbelRDistribution
from .half_cauchy import HalfCauchyDistribution
from .half_generalized_normal import HalfGeneralizedNormalDistribution
from .half_logistic import HalfLogisticDistribution
from .half_normal import HalfNormalDistribution
from .hyperbolic_secant import HyperbolicSecantDistribution
from .inverse_gamma import InverseGammaDistribution
from .inverse_weibull import InverseWeibullDistribution
from .johnson_sb import JohnsonSBDistribution
from .johnson_su import JohnsonSUDistribution
from .kappa_3 import Kappa3Distribution
from .kappa_4 import Kappa4Distribution
from .ks_one import KSOneDistribution
from .ks_two_bi_generalized import KSTwoBinomialGeneralizedDistribution
from .ks_two import KSTwoDistribution
from .laplace_asymmetric import LaplaceAsymmetricDistribution
from .laplace import LaplaceDistribution
from .levy_l import LevyLDistribution
from .levy_stable import LevyStableDistribution
from .levy import LevyDistribution
from .log_gamma import LogGammaDistribution
from .log_laplace import LogLaplaceDistribution
from .log_uniform import LogUniformDistribution
from .logistic import LogisticDistribution
from .lomax import LomaxDistribution
from .maxwell import MaxwellDistribution
from .mielke import MielkeDistribution
from .moyal import MoyalDistribution
from .nakagami import NakagamiDistribution
from .non_central_f_distribution import NonCenteralFDistribution
from .noncentral_chi_squared import NonCenteralChiSquaredDistribution
from .noncentral_t_distribution import NonCenteralTDistribution
from .normal_inverse_gauss import NormalInverseGaussDistribution
from .normal import NormalDistribution
from .pareto import ParetoDistribution
from .pearson_3 import Pearson3Distribution
from .power_log_normal import PowerLogNormalDistribution
from .power_normal import PowerNormalDistribution
from .powerlaw import PowerlawDistribution
from .r_distribution import RDistribution
from .rayleigh import RayleighDistribution
from .reciprocal_inverse_gauss import ReciprocalInverseGaussDistribution
from .rice import RiceDistribution
from .semi_circular import SemiCircularDistribution
from .skewed_cauchy import SkewedCauchyDistribution
from .skewed_normal import SkewedNormalDistribution
from .student_range import StudentRangeDistribution
from .t_distribution import TDistribution
from .trapezoid import TrapezoidDistribution
from .triangular import TriangularDistribution
from .truncated_exponential import TruncatedExponentialDistribution
from .truncated_normal import TruncatedNormalDistribution
from .truncated_pareto import TruncatedParetoDistribution
from .truncated_weibull_minimum import TruncatedWeibullMinimumDistribution
from .tukey_lambda import TukeyLambdaDistribution
from .uniform import UniformDistribution
from .vonmises_line import VonMisesLineDistribution
from .vonmises import VonMisesDistribution
from .wald import WaldDistribution
from .weibull_maximum import WeibullMaxmimumDistribution
from .weibull_minimum import WeibullMinimumDistribution
from .wrapped_cauchy import WrappedCauchyDistribution



from pydantic import BaseModel, StrictStr
from hedra.data.connectors.common.connector_type import ConnectorType


class JSONConnectorConfig(BaseModel):
    filepath: StrictStr
    file_mode: StrictStr='r'
    connector_type: ConnectorType=ConnectorType.JSON


from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.task.task import Task
from hedra.core.engines.types.task.runner import MercuryTaskRunner
from hedra.core.engines.types.task.result import TaskResult
from hedra.core.engines.types.common.types import RequestTypes
from hedra.data.serializers.serializer_types.common.base_serializer import BaseSerializer
from typing import List, Dict, Union, Any


class TaskSerializer(BaseSerializer):

    def __init__(self) -> None:
        super().__init__()

    def action_to_serializable(
        self,
        task: Task
    ) -> Dict[str, Union[str, List[str]]]:
        
        serialized_action = super().action_to_serializable(task)
        return {
            **serialized_action,
            'type': RequestTypes.TASK,
            'source': task.source,
            'task_action': task.execute.__name__,
            'user': task.metadata.user,
            'tags': task.metadata.tags
        }
    
    def deserialize_task(
        self,
        task: Dict[str, Any]
    ) -> Task:
        
        deserialized_task = Task(
            task.get('name'),
            None,
            source=task.get('source'),
            user=task.get('user'),
            tags=task.get('tags', [])
        )

        return deserialized_task
    
    def deserialize_client_config(self, client_config: Dict[str, Any]) -> MercuryTaskRunner:
        return MercuryTaskRunner(
            concurrency=client_config.get('concurrency'),
            timeouts=Timeouts(
                **client_config.get('timeouts', {})
            )
        )
    
    def result_to_serializable(
        self,
        result: TaskResult
    ) -> Dict[str, Any]:
    
        serialized_result = super().result_to_serializable(result)
        return {
            **serialized_result,
            'data': result.data,
        }
    
    def deserialize_result(
        self,
        result: Dict[str, Any]
    ) -> TaskResult:
        
        task_action = Task(
            result.get('name'),
            None,
            source=result.get('source'),
            user=result.get('user'),
            tags=result.get('tags')
        )

        task_result = TaskResult(
            task_action,
            error=result.get('error')
        )

        task_result.data = result.get('data')
     
  
        task_result.checks = result.get('checks')
        task_result.wait_start = result.get('wait_start')
        task_result.start = result.get('start')
        task_result.connect_end = result.get('connect_end')
        task_result.write_end = result.get('write_end')
        task_result.complete = result.get('complete')

        return task_result



from typing import (
    Coroutine, 
    Dict, 
    Any, 
    Callable, 
    Awaitable, 
    Tuple
)
from hedra.core.hooks.types.base.hook_type import HookType
from hedra.core.hooks.types.base.hook import Hook


class ConditionHook(Hook):

    def __init__(
        self, 
        name: str, 
        shortname: str, 
        call: Callable[..., Awaitable[Any]], 
        *names: Tuple[str, ...],
        order: int=1,
        skip: bool=False
    ) -> None:
        super().__init__(
            name, 
            shortname, 
            call, 
            order=order,
            skip=skip,
            hook_type=HookType.CONDITION
        )

        self.names = list(set(names))
        self.pre: bool = True
        self.key: str = None
        self.events: Dict[str, Coroutine] = {}

    async def call(self, **kwargs):

        if self.skip:
            return kwargs

        execute = await super().call(**{name: value for name, value in kwargs.items() if name in self.params})

        if isinstance(execute, dict):
            return {
                **kwargs,
                **execute
            }

        return {
            **kwargs,
            self.shortname: execute
        }
    
    def copy(self):
        condition_hook = ConditionHook(
            self.name,
            self.shortname,
            self._call,
            *self.names,
            order=self.order,
            skip=self.skip,
        )

        condition_hook.stage = self.stage

        return condition_hook


import asyncio
import uuid
from typing import (
    List, 
    Dict, 
    Union,
    Any,
    Coroutine
)
from hedra.logging import HedraLogger
from hedra.core.engines.client.config import Config
from hedra.core.engines.types.common.results_set import ResultsSet
from hedra.core.hooks.types.action.hook import ActionHook
from hedra.data.connectors.common.connector_type import ConnectorType
from hedra.data.connectors.common.execute_stage_summary_validator import ExecuteStageSummaryValidator
from hedra.data.parsers.parser import Parser
from .sqlite_connector_config import SQLiteConnectorConfig


def noop_create_async_engine():
    pass


try:
    import sqlalchemy
    from sqlalchemy.engine.result import Result as SQLResult
    from sqlalchemy.ext.asyncio import (
        create_async_engine,
        AsyncEngine,
        AsyncConnection,
    )
        
    has_connector = True

except Exception:
    sqlalchemy = object
    SQLResult = None
    create_async_engine = noop_create_async_engine
    AsyncEngine = object
    AsyncConnection = object
    has_connector = False



class SQLiteConnector:
    connector_type=ConnectorType.SQLite

    def __init__(
        self, 
        config: SQLiteConnectorConfig,
        stage: str,
        parser_config: Config,
    ) -> None:
        self.path = f'sqlite+aiosqlite:///{config.path}'
        self.table_name = config.table_name
        self.stage = stage
        self.parser_config = parser_config

        self.metadata = sqlalchemy.MetaData()

        self.database = None
        self._engine: Union[AsyncEngine, None] = None
        self._connection: Union[AsyncConnection, None] = None

        self._table: Union[sqlalchemy.Table, None] = None

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.parser = Parser()

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to SQLite at - {self.path} - Database: {self.database}')
        self._engine = create_async_engine(self.path)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to SQLite at - {self.path} - Database: {self.database}')

    async def load_execute_stage_summary(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ExecuteStageSummaryValidator]:
        execute_stage_summary = await self.load_data(
            options=options
        )
        
        return ExecuteStageSummaryValidator(**execute_stage_summary)

    async def load_actions(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[ActionHook]]:
        actions = await self.load_data(
            options=options
        )

        return await asyncio.gather(*[
            self.parser.parse_action(
                action_data,
                self.stage,
                self.parser_config,
                options
            ) for action_data in actions
        ])
    
    async def load_results(
        self,
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, ResultsSet]:
        results = await self.load_data(
            options=options
        )

        return ResultsSet({
            'stage_results': await asyncio.gather(*[
                self.parser.parse_result(
                    results_data,
                    self.stage,
                    self.parser_config,
                    options
                ) for results_data in results
            ])
        })
    
    async def load_data(
        self, 
        options: Dict[str, Any]={}
    ) -> Coroutine[Any, Any, List[Dict[str, Any]]]:
        
        if self._table is None:
            self._table = sqlalchemy.Table(
                self.table_name,
                self.metadata,
                autoload_with=self._engine
            )

        async with self._engine.connect() as connection:
            connection: AsyncConnection = connection
            results: SQLResult = await self._connection.execute(
                self._table.select(**options)
            )
            
            return [
                {
                    column: value  for column, value in result._mapping.items()
                } for result in results.fetchall()
            ]
    
    async def close(self):
        await self._connection.close()


from typing import Dict, Callable, List, Type
from hedra.versioning.flags.types.unstable.flag import unstable
from .distribution_types import (
    DistributionTypes,
    DistributionMap
)
from .distributions.types import (
    AlphaDistribution,
    AnglitDistribution,
    ArcsineDistribution,
    ArgusDistribution,
    BetaDistribution,
    BetaPrimeDistribution,
    BradfordDistribution,
    Burr12Distribution,
    BurrDistribution,
    CauchyDistribution,
    ChiDistribution,
    ChiSquaredDistribution,
    CosineDistribution,
    CrystalDistribution,
    DGammaDistribution,
    DWeibullDistribution,
    ErlangDistribution,
    ExponentialDistribution,
    ExponentialNormalDistribution,
    ExponentialPowerDistribution,
    FatigueLifeDistribution,
    FDistribution,
    FiskDistribution,
    FoldedCauchyDistribution,
    FoldedNormalDistribution,
    GammaDistribution,
    GaussHypergeometricDistribution,
    GeneralizedExponentialDistribution,
    GeneralizedExtremeDistribution,
    GeneralizedGammaDistribution,
    GeneralizedHalfLogisticDistribution,
    GeneralizedHyperbolicDistribution,
    GeneralizedInverseGaussDistribution,
    GeneralizedLogisticDistribution,
    GeneralizedNormalDistribution,
    GeneralizedParetoDistribution,
    GibratDistribution,
    GompertzDistribution,
    GumbelLDistribution,
    GumbelRDistribution,
    HalfCauchyDistribution,
    HalfGeneralizedNormalDistribution,
    HalfLogisticDistribution,
    HalfNormalDistribution,
    HyperbolicSecantDistribution,
    InverseGammaDistribution,
    InverseWeibullDistribution,
    JohnsonSBDistribution,
    JohnsonSUDistribution,
    Kappa3Distribution,
    Kappa4Distribution,
    KSOneDistribution,
    KSTwoBinomialGeneralizedDistribution,
    KSTwoDistribution,
    LaplaceAsymmetricDistribution,
    LaplaceDistribution,
    LevyDistribution,
    LevyLDistribution,
    LevyStableDistribution,
    LogGammaDistribution,
    LogisticDistribution,
    LogLaplaceDistribution,
    LogUniformDistribution,
    LomaxDistribution,
    MaxwellDistribution,
    MielkeDistribution,
    MoyalDistribution,
    NakagamiDistribution,
    NonCenteralChiSquaredDistribution,
    NonCenteralFDistribution,
    NonCenteralTDistribution,
    NormalDistribution,
    NormalInverseGaussDistribution,
    ParetoDistribution,
    Pearson3Distribution,
    PowerlawDistribution,
    PowerLogNormalDistribution,
    PowerNormalDistribution,
    RayleighDistribution,
    RDistribution,
    ReciprocalInverseGaussDistribution,
    RiceDistribution,
    SemiCircularDistribution,
    SkewedCauchyDistribution,
    SkewedNormalDistribution,
    StudentRangeDistribution,
    TDistribution,
    TrapezoidDistribution,
    TriangularDistribution,
    TruncatedExponentialDistribution,
    TruncatedNormalDistribution,
    TruncatedParetoDistribution,
    TruncatedWeibullMinimumDistribution,
    TukeyLambdaDistribution,
    UniformDistribution,
    VonMisesDistribution,
    VonMisesLineDistribution,
    WaldDistribution,
    WeibullMaxmimumDistribution,
    WeibullMinimumDistribution,
    WrappedCauchyDistribution
)
from .distributions.types.base import BaseDistribution


@unstable
class Distribution:

    def __init__(
            self,  
            distribution_type: DistributionTypes=DistributionTypes.NORMAL,
            intervals: int=None
        ) -> None:

        self._distribution_map = DistributionMap()
        self._distributions: Dict[str, Callable[..., List[float]]] = {
            DistributionTypes.ALPHA: lambda size: AlphaDistribution(size),
            DistributionTypes.ANGLIT: lambda size: AnglitDistribution(size),
            DistributionTypes.ARCSINE: lambda size: ArcsineDistribution(size),
            DistributionTypes.ARGUS: lambda size: ArgusDistribution(size),
            DistributionTypes.BETA_PRIME: lambda size: BetaPrimeDistribution(size),
            DistributionTypes.BETA: lambda size: BetaDistribution(size),
            DistributionTypes.BRADFORD: lambda size: BradfordDistribution(size),
            DistributionTypes.BURR: lambda size: BurrDistribution(size),
            DistributionTypes.BURR12: lambda size: Burr12Distribution(size),
            DistributionTypes.CAUCHY: lambda size: CauchyDistribution(size),
            DistributionTypes.CHI_SQUARED: lambda size: ChiSquaredDistribution(size),
            DistributionTypes.CHI: lambda size: ChiDistribution(size),
            DistributionTypes.COSINE: lambda size: CosineDistribution(size),
            DistributionTypes.CRYSTAL_BALL: lambda size: CrystalDistribution(size),
            DistributionTypes.DGAMMA: lambda size: DGammaDistribution(size),
            DistributionTypes.DWEIBULL: lambda size: DWeibullDistribution(size),
            DistributionTypes.ERLANG: lambda size: ErlangDistribution(size),
            DistributionTypes.EXPONENTIAL_NORMAL: lambda size: ExponentialNormalDistribution(size),
            DistributionTypes.EXPONENTIAL_POWER: lambda size: ExponentialPowerDistribution(size),
            DistributionTypes.EXPONENTIAL: lambda size: ExponentialDistribution(size),
            DistributionTypes.F_DISTRIBUTION: lambda size: FDistribution(size),
            DistributionTypes.FATIGUE_LIFE: lambda size: FatigueLifeDistribution(size),
            DistributionTypes.FISK: lambda size: FiskDistribution(size),
            DistributionTypes.FOLDED_CAUCHY: lambda size: FoldedCauchyDistribution(size),
            DistributionTypes.FOLDED_NORMAL: lambda size: FoldedNormalDistribution(size),
            DistributionTypes.GAMMA: lambda size: GammaDistribution(size),
            DistributionTypes.GAUSS_HYPERGEOMETRIC: lambda size: GaussHypergeometricDistribution(size),
            DistributionTypes.GENERALIZED_EXPONENTIAL: lambda size: GeneralizedExponentialDistribution(size),
            DistributionTypes.GENERALIZED_EXTREME: lambda size: GeneralizedExtremeDistribution(size),
            DistributionTypes.GENERALIZED_GAMMA: lambda size: GeneralizedGammaDistribution(size),
            DistributionTypes.GENERALIZED_HALF_LOGISTIC: lambda size: GeneralizedHalfLogisticDistribution(size),
            DistributionTypes.GENERALIZED_HYPERBOLIC: lambda size: GeneralizedHyperbolicDistribution(size),
            DistributionTypes.GENERALIZED_INVERSE_GAUSS: lambda size: GeneralizedInverseGaussDistribution(size),
            DistributionTypes.GENERALIZED_LOGISTIC: lambda size: GeneralizedLogisticDistribution(size),
            DistributionTypes.GENERALIZED_NORMAL: lambda size: GeneralizedNormalDistribution(size),
            DistributionTypes.GENERALIZED_PARETO: lambda size: GeneralizedParetoDistribution(size),
            DistributionTypes.GIBRAT: lambda size: GibratDistribution(size),
            DistributionTypes.GOMPERTZ: lambda size: GompertzDistribution(size),
            DistributionTypes.GUMBEL_L: lambda size: GumbelLDistribution(size),
            DistributionTypes.GUMBEL_R: lambda size: GumbelRDistribution(size),
            DistributionTypes.HALF_CAUCHY: lambda size: HalfCauchyDistribution(size),
            DistributionTypes.HALF_GENERALIZED_NORMAL: lambda size: HalfGeneralizedNormalDistribution(size),
            DistributionTypes.HALF_LOGISTIC: lambda size: HalfLogisticDistribution(size),
            DistributionTypes.HALF_NORMAL: lambda size: HalfNormalDistribution(size),
            DistributionTypes.HYPERBOLIC_SECANT: lambda size: HyperbolicSecantDistribution(size),
            DistributionTypes.INVERSE_GAMMA: lambda size: InverseGammaDistribution(size),
            DistributionTypes.INVERSE_WEIBULL: lambda size: InverseWeibullDistribution(size),
            DistributionTypes.JOHNSON_SB: lambda size: JohnsonSBDistribution(size),
            DistributionTypes.JOHNSON_SU: lambda size: JohnsonSUDistribution(size),
            DistributionTypes.KAPPA_3: lambda size: Kappa3Distribution(size),
            DistributionTypes.KAPPA_4: lambda size: Kappa4Distribution(size),
            DistributionTypes.KS_ONE: lambda size: KSOneDistribution(size),
            DistributionTypes.KS_TWO: lambda size: KSTwoDistribution(size),
            DistributionTypes.KS_TWO_BIGENERALIZED: lambda size: KSTwoBinomialGeneralizedDistribution(size),
            DistributionTypes.LAPLACE: lambda size: LaplaceDistribution(size),
            DistributionTypes.LAPLACE_ASYMMETRIC: lambda size: LaplaceAsymmetricDistribution(size),
            DistributionTypes.LEVY: lambda size: LevyDistribution(size),
            DistributionTypes.LEVY_L: lambda size: LevyLDistribution(size),
            DistributionTypes.LEVY_STABLE: lambda size: LevyStableDistribution(size),
            DistributionTypes.LOG_GAMMA: lambda size: LogGammaDistribution(size),
            DistributionTypes.LOG_LAPLAPCE: lambda size: LogLaplaceDistribution(size),
            DistributionTypes.LOG_UNIFORM: lambda size: LogUniformDistribution(size),
            DistributionTypes.LOGISTIC: lambda size: LogisticDistribution(size),
            DistributionTypes.LOMAX: lambda size: LomaxDistribution(size),
            DistributionTypes.MAXWELL: lambda size: MaxwellDistribution(size),
            DistributionTypes.MIELKE: lambda size: MielkeDistribution(size),
            DistributionTypes.MOYAL: lambda size: MoyalDistribution(size),
            DistributionTypes.NAKAGAMI: lambda size: NakagamiDistribution(size),
            DistributionTypes.NON_CENTRAL_CHI_SQUARED: lambda size: NonCenteralChiSquaredDistribution(size),
            DistributionTypes.NON_CENTRAL_F_DISTRIBUTION: lambda size: NonCenteralFDistribution(size),
            DistributionTypes.NON_CENTRAL_T_DISTRIBUTION: lambda size: NonCenteralTDistribution(size),
            DistributionTypes.NORMAL_INVERSE_GAUSS: lambda size: NormalInverseGaussDistribution(size),
            DistributionTypes.NORMAL: lambda size: NormalDistribution(size),
            DistributionTypes.PARETO: lambda size: ParetoDistribution(size),
            DistributionTypes.PEARSON_3: lambda size: Pearson3Distribution(size),
            DistributionTypes.POWER_LOG_NORMAL: lambda size: PowerLogNormalDistribution(size),
            DistributionTypes.POWER_NORMAL: lambda size: PowerNormalDistribution(size),
            DistributionTypes.POWERLAW: lambda size: PowerlawDistribution(size),
            DistributionTypes.R_DISTRIBUTION: lambda size: RDistribution(size),
            DistributionTypes.RAYLEIGH: lambda size: RayleighDistribution(size),
            DistributionTypes.RECIPROCAL_INVERSE_GAUSS: lambda size: ReciprocalInverseGaussDistribution(size),
            DistributionTypes.RICE: lambda size: RiceDistribution(size),
            DistributionTypes.SEMI_CIRCULAR: lambda size: SemiCircularDistribution(size),
            DistributionTypes.SKEWED_CAUCHY: lambda size: SkewedCauchyDistribution(size),
            DistributionTypes.SKEWED_NORMAL: lambda size: SkewedNormalDistribution(size),
            DistributionTypes.STUDENT_RANGE: lambda size: StudentRangeDistribution(size),
            DistributionTypes.T_DISTRIBUTION: lambda size: TDistribution(size),
            DistributionTypes.TRAPEZOID: lambda size: TrapezoidDistribution(size),
            DistributionTypes.TRIANGULAR: lambda size: TriangularDistribution(size),
            DistributionTypes.TRUNCATED_EXPONENTIAL: lambda size: TruncatedExponentialDistribution(size),
            DistributionTypes.TRUNCATED_NORMAL: lambda size: TruncatedNormalDistribution(size),
            DistributionTypes.TRUNCATED_PARETO: lambda size: TruncatedParetoDistribution(size),
            DistributionTypes.TRUNCATED_WEIBULL_MINIMUM: lambda size: TruncatedWeibullMinimumDistribution(size),
            DistributionTypes.TUKEY_LAMBDA: lambda size: TukeyLambdaDistribution(size),
            DistributionTypes.UNIFORM: lambda size: UniformDistribution(size),
            DistributionTypes.VONMISES_LINE: lambda size: VonMisesLineDistribution(size),
            DistributionTypes.VONMISES: lambda size: VonMisesDistribution(size),
            DistributionTypes.WALD: lambda size: WaldDistribution(size),
            DistributionTypes.WEIBULL_MAXIMUM: lambda size: WeibullMaxmimumDistribution(size),
            DistributionTypes.WEIBULL_MINIMUM: lambda size: WeibullMinimumDistribution(size),
            DistributionTypes.WRAPPED_CAUCHY: lambda size: WrappedCauchyDistribution(size)
        }

        self.selected_distribution: DistributionTypes = self._distribution_map.get(
            distribution_type
        )
        self.disribution_function: Type[BaseDistribution] = self._distributions.get(
            self.selected_distribution
        )

        self.intervals = intervals

    def generate(self, batch_size: int):
        distribution = self.disribution_function(self.intervals)
        return distribution.generate_distribution(batch_size)

    


from typing import (
    Dict,
    List,
    Iterable,
    Union
)
from hedra.core.engines.types.graphql.action import GraphQLAction
from hedra.core.engines.types.graphql_http2.action import GraphQLHTTP2Action
from hedra.core.engines.types.grpc.action import GRPCAction
from hedra.core.engines.types.http.action import HTTPAction
from hedra.core.engines.types.http2.action import HTTP2Action
from hedra.core.engines.types.http3.action import HTTP3Action
from hedra.core.engines.types.playwright.command import PlaywrightCommand
from hedra.core.engines.types.task.task import Task
from hedra.core.engines.types.udp.action import UDPAction
from hedra.core.engines.types.websocket.action import WebsocketAction


Action = Union[
    GraphQLAction,
    GraphQLHTTP2Action,
    GRPCAction,
    HTTPAction,
    HTTP2Action,
    HTTP3Action,
    PlaywrightCommand,
    Task,
    UDPAction,
    WebsocketAction
]


class ActionRegistry:

    def __init__(self) -> None:
        self._actions: Dict[str, Action] = {}

    def __iter__(self) -> Iterable[Action]:
        for action in self._actions.values():
            yield action

    def __setitem__(self, name: str, value: Action):
        self._actions[name] = value

    def actions(self) -> List[Action]:
        return list(self._actions.values())

    def names(self):
        for action_name in self._actions.keys():
            yield action_name

    def get_action(self, action_name: str) -> Union[Action, None]:
        return self._actions.get(action_name)

    def get_stage_actions(self, stage_name: str) -> List[Action]:
        return [
            action for action in self._actions.values() if action.stage == stage_name
        ]
            

def make_action_registry():
    return ActionRegistry()


actions_registry = make_action_registry()


import os
from pydantic import (
    BaseModel,
    StrictStr
)

from hedra.data.connectors.common.connector_type import ConnectorType


class SQLiteConnectorConfig(BaseModel):
    path: StrictStr=f'{os.getcwd()}/results.db'
    table_name:StrictStr
    reporter_type: ConnectorType=ConnectorType.SQLite

    class Config:
        arbitrary_types_allowed = True